@@ -132,7 +132,7 @@ type put_copy_result =
 
 (** Result of get_copy_data *)
 type get_copy_result =
-  | Get_copy_data of string (** Data corresponding to one row is retured *)
+  | Get_copy_data of string (** Data corresponding to one row is returned *)
   | Get_copy_wait   (** The next row is still being received (async only); wait
                         for read-only, call [consume_input], and try again *)
   | Get_copy_end    (** All data has been successfully retrieved *)
@@ -205,7 +205,7 @@ val string_of_ftype : ftype -> string
 (** [string_of_ftype ftype] converts [ftype] to a string. *)
 
 val ftype_of_string : string -> ftype
-(** [string_of_ftype ftype] converts [ftype] to a string. *)
+(** [string_of_ftype ftype] converts string to a [ftype]. *)
 
 
 (** {2 Handling results of commands and queries} *)
@@ -586,6 +586,18 @@ object
   (** [#set_notice_processor] controls reporting of notice and warning
       messages generated by a connection.
 
+      {e Warning:} This function is unsafe in combination with a number of libpq
+      entry points, and should not be used for now. As a workaround,
+      {!#set_notice_processing} can be used to silence notices, if this is more
+      appropriate than the default behaviour of printing them to standard error.
+
+      @raise Error if there is a connection error.
+  *)
+
+  method set_notice_processing : [`Stderr | `Quiet] -> unit
+  (** [#set_notice_processing] controls reporting of notice and warning messages
+      generated by a connection by providing predefined callbacks.
+
       @raise Error if there is a connection error.
   *)
 
