@@ -25,11 +25,9 @@
 */
 
 #if __GNUC__ >= 3
-# define inline inline __attribute__ ((always_inline))
 # define __unused __attribute__ ((unused))
 #else
 # define __unused
-# define inline
 #endif
 
 #if PG_OCAML_MAJOR_VERSION > 8 \
@@ -132,8 +130,8 @@ static inline value make_some(value v)
 
 /* Cache for OCaml-values */
 static value v_empty_string = Val_unit;
-static value *v_exc_Oid = NULL;  /* Exception [Oid] */
-static value *v_null_param = NULL;
+static const value *v_exc_Oid = NULL;  /* Exception [Oid] */
+static const value *v_null_param = NULL;
 
 CAMLprim value PQocaml_init(value __unused v_unit)
 {
@@ -528,13 +526,23 @@ static struct custom_operations result_ops = {
   custom_hash_default,
   custom_serialize_default,
   custom_deserialize_default,
-  custom_compare_ext_default
+  custom_compare_ext_default,
+  custom_fixed_length_default
 };
 
 static inline value alloc_result(PGresult *res, np_callback *cb)
 {
-  value v_res =
-    caml_alloc_custom(&result_ops, sizeof(struct pg_ocaml_result), 1, 100000);
+  value v_res;
+  size_t result_size, ocaml_result_size = sizeof(struct pg_ocaml_result);
+#if PG_OCAML_MAJOR_VERSION < 12
+  /* There isn't really a bound on size here, it all depends on the number
+     of records, columns, and type of data.  4096 bytes seems like a reasonable
+     size that shouldn't lead to excessive GC pressure. */
+  result_size = 4096;
+#else
+  result_size = PQresultMemorySize(res);
+#endif
+  v_res = caml_alloc_custom_mem(&result_ops, ocaml_result_size, result_size);
   set_res(v_res, res);
   set_res_cb(v_res, cb);
   np_incr_refcount(cb);
@@ -604,7 +612,7 @@ static inline void free_params(const char * const *params, size_t nparams)
 static inline const char * const * copy_params_shallow(
   value v_params, size_t nparams)
 {
-  char **params;
+  const char **params;
   size_t i;
   if (nparams == 0) return NULL;
   params = caml_stat_alloc(nparams * sizeof(char *));
@@ -807,8 +815,7 @@ CAMLprim value PQgetvalue_stub(value v_res, intnat tup_num, intnat field_num)
   else {
     /* Assume binary format! */
     size_t len = PQgetlength(res, tup_num, field_num);
-    v_str = len ? caml_alloc_string(len) : v_empty_string;
-    memcpy(String_val(v_str), str, len);
+    v_str = len ? caml_alloc_initialized_string(len, str) : v_empty_string;
   }
   CAMLreturn(v_str);
 }
@@ -862,8 +869,7 @@ static value unescape_bytea(const char *str)
   char *buf = (char *) PQunescapeBytea((unsigned char*) str, &res_len);
   if (buf == NULL) caml_failwith("Postgresql: illegal bytea string");
   else {
-    value v_res = caml_alloc_string(res_len);
-    memcpy(String_val(v_res), buf, res_len);
+    value v_res = caml_alloc_initialized_string(res_len, buf);
     PQfreemem(buf);
     return v_res;
   }
@@ -909,13 +915,12 @@ CAMLprim value PQgetescval_stub(value v_res, intnat tup_num, intnat field_num)
       str += 2;
       n_hex_pairs = bytea_hex_pairs(str);
       v_str = caml_alloc_string(n_hex_pairs);
-      decode_bytea_hex(str, String_val(v_str), n_hex_pairs);
+      decode_bytea_hex(str, (char *) String_val(v_str), n_hex_pairs);
     }
   } else {
     /* Assume binary format! */
     size_t len = PQgetlength(res, tup_num, field_num);
-    v_str = len ? caml_alloc_string(len) : v_empty_string;
-    memcpy(String_val(v_str), str, len);
+    v_str = len ? caml_alloc_initialized_string(len, str) : v_empty_string;
   }
   CAMLreturn(v_str);
 }
@@ -1094,10 +1099,21 @@ CAMLprim value PQgetResult_stub(value v_conn)
 }
 
 noalloc_conn_info_intnat(PQconsumeInput)
-noalloc_conn_info(PQisBusy, Val_bool)
+
 noalloc_conn_info_intnat(PQflush)
 noalloc_conn_info_intnat(PQsocket)
 
+CAMLprim value PQisBusy_stub(value v_conn)
+{
+  CAMLparam1(v_conn);
+  PGconn *conn = get_conn(v_conn);
+  bool res;
+  caml_enter_blocking_section();
+    res = PQisBusy(conn);
+  caml_leave_blocking_section();
+  CAMLreturn(Val_bool(res));
+}
+
 CAMLprim value PQCancel_stub(value v_conn)
 {
   CAMLparam1(v_conn);
@@ -1110,8 +1126,8 @@ CAMLprim value PQCancel_stub(value v_conn)
     caml_enter_blocking_section();
       res = PQcancel(cancel, errbuf, 256);
     caml_leave_blocking_section();
-    if (res == 0) CAMLreturn(v_None);
-    else CAMLreturn(make_some(caml_copy_string(errbuf)));
+    if (res == 0) CAMLreturn(make_some(caml_copy_string(errbuf)));
+    else CAMLreturn(v_None);
   }
 }
 
@@ -1128,8 +1144,7 @@ CAMLprim value PQescapeStringConn_stub(
     caml_stat_free(buf);
     caml_failwith("Postgresql.escape_string_conn: failed to escape string");
   } else {
-    value v_res = caml_alloc_string(n_written);
-    memcpy(String_val(v_res), buf, n_written);
+    value v_res = caml_alloc_initialized_string(n_written, buf);
     caml_stat_free(buf);
     return v_res;
   }
@@ -1151,8 +1166,7 @@ CAMLprim value PQescapeByteaConn_stub(
     (char *) PQescapeByteaConn(
       get_conn(v_conn),
       (unsigned char *) String_val(v_from) + pos_from, len, &res_len);
-  value v_res = caml_alloc_string(--res_len);
-  memcpy(String_val(v_res), buf, res_len);
+  value v_res = caml_alloc_initialized_string(--res_len, buf);
   PQfreemem(buf);
   return v_res;
 }
@@ -1177,7 +1191,8 @@ CAMLprim value PQunescapeBytea_stub(value v_from)
     CAMLparam1(v_from);
     value v_res = caml_alloc_string(res_len);
     /* GC may have happened, have to use String_val on v_from again */
-    decode_bytea_hex(String_val(v_from) + 2, String_val(v_res), res_len);
+    decode_bytea_hex(
+      String_val(v_from) + 2, (char *) String_val(v_res), res_len);
     CAMLreturn(v_res);
   }
 }
@@ -1277,8 +1292,7 @@ CAMLprim value PQgetCopyData_stub(value v_conn, intnat async)
     case -2:
       CAMLreturn(Val_int(2)); /* Get_copy_error */
     default:
-      v_buf = caml_alloc_string(res);
-      memcpy(String_val(v_buf), buf, res);
+      v_buf = caml_alloc_initialized_string(res, buf);
       PQfreemem(buf);
       v_res = caml_alloc_small(1, 0); /* Get_copy_data */
       Field(v_res, 0) = v_buf;
@@ -1301,7 +1315,7 @@ CAMLprim intnat PQgetline_stub(
   caml_enter_blocking_section();
     res = PQgetline(conn, buf, len);
   caml_leave_blocking_section();
-  memcpy(String_val(v_buf) + pos, buf, len);
+  memcpy(Bytes_val(v_buf) + pos, buf, len);
   caml_stat_free(buf);
   CAMLreturn(res);
 }
@@ -1315,7 +1329,7 @@ CAMLprim value PQgetline_stub_bc(
 CAMLprim intnat PQgetlineAsync_stub(
   value v_conn, value v_buf, intnat pos, intnat len)
 {
-  return PQgetlineAsync(get_conn(v_conn), String_val(v_buf) + pos, len);
+  return PQgetlineAsync(get_conn(v_conn), (char *) Bytes_val(v_buf) + pos, len);
 }
 
 CAMLprim value PQgetlineAsync_stub_bc(
@@ -1388,9 +1402,9 @@ CAMLprim value PQendcopy_stub_bc(value v_conn)
 static inline void notice_ml(void *cb, const char *msg)
 {
   value v_msg;
-  /* CR mmottl for mmottl: this is not reliable and can lead to segfaults,
-     because the runtime lock may already be held (but not usually).
-     A runtime feature is needed to fully support this. */
+  /* CR mmottl for mmottl: this is not reliable and can lead to deadlocks or
+     other unintended behavior, because the runtime lock may already be held
+     (but not usually).  A runtime feature is needed to fully support this. */
   caml_leave_blocking_section();
     v_msg = make_string(msg);
     caml_callback(((np_callback *) cb)->v_cb, v_msg);
@@ -1405,6 +1419,31 @@ CAMLprim value PQsetNoticeProcessor_stub(value v_conn, value v_cb)
   return Val_unit;
 }
 
+static void np_quiet(void __unused *arg, const char __unused *message)
+{
+}
+
+static void np_stderr(void __unused *arg, const char *message)
+{
+    fprintf(stderr, "%s", message);
+}
+
+CAMLprim value PQsetNoticeProcessor_num(value v_conn, value v_cb_num)
+{
+  np_decr_refcount(get_conn_cb(v_conn));
+  set_conn_cb(v_conn, NULL);
+  switch (Int_val(v_cb_num)) {
+    case 0:
+      PQsetNoticeProcessor(get_conn(v_conn), np_stderr, NULL);
+      break;
+    case 1:
+      PQsetNoticeProcessor(get_conn(v_conn), np_quiet, NULL);
+      break;
+    default:
+      break;
+  }
+  return Val_unit;
+}
 
 /* Large objects */
 
@@ -1538,7 +1577,7 @@ CAMLprim intnat lo_read_stub(
   caml_enter_blocking_section();
     res = lo_read(conn, fd, buf, len);
   caml_leave_blocking_section();
-  memcpy(String_val(v_buf) + pos, buf, len);
+  memcpy(Bytes_val(v_buf) + pos, buf, len);
   caml_stat_free(buf);
   CAMLreturn(res);
 }
